#!/usr/bin/env python3
from pydantic import BaseSettings


class ServiceContext:
    def __init__(
            self,
            name: str,
            settings: BaseSettings,
    ):
        self.name = name
        self.settings = settings



import asyncio
import signal


from .importer import import_from_string
from .service import Service

HANDLED_SIGNALS = (
    signal.SIGINT,  # Unix signal 2. Sent by Ctrl+C.
    signal.SIGTERM,  # Unix signal 15. Sent by `kill <pid>`.
)

class ServiceRunner:
    def __init__(
            self,
            *,
            service_str,
    ):
        self.service_str = service_str


    def load(self):
        self.loaded_service: Service = import_from_string(self.service_str)

    def run(self):
        self.load()
        return asyncio.run(self.loaded_service.run())



__version__ = "0.1.1"


from uservice.main import main
from uservice.service import Service
from uservice.amqp.events import AmqpEventPublisher as EventPublisher
from uservice.amqp.rpc import AmqpRpcProxy as RpcProxy

__all__ = [
    "main",
    "Service",
    "EventPublisher",
    "RpcProxy",
]



import asyncio
import json
from aio_pika import connect, Message

from .main import Test


async def main():
    connection = await connect()
    async with connection:
        channel = await connection.channel()
        exchange = await channel.get_exchange('test')
        print("ello")
        for i in range(1):
            await exchange.publish(
                Message(body=json.dumps(Test(foo=i).dict()).encode()),
                routing_key='event'
            )



if __name__ == '__main__':
    asyncio.run(main())



from inspect import isclass
from contextlib import AsyncExitStack
from typing import Callable, Any, AsyncGenerator

from aio_pika import IncomingMessage, Connection, Message
from pydantic import BaseModel, ValidationError

from uservice.contexts import ServiceContext
from uservice.utils import create_publish_field, serialize_payload
from .consumer import AmqpConsumer


class AmqpEventHandler(AmqpConsumer):
    def __init__(
            self,
            *,
            context: ServiceContext,
            call: Callable,
            exchange_name: str,
            routing_key: str,
    ):
        super().__init__(context=context, call=call)
        self.exchange_name = exchange_name
        self.routing_key = routing_key
        self
        try:
            self.payload_type = self.dependant.required_params['payload'].annotation
        except KeyError:
            raise ValueError('Event handler function is missing required arugment "payload"')

    def get_exchange_name(self) -> str:
        return self.exchange_name

    def get_routing_key(self) -> str:
        return self.routing_key

    def get_queue_name(self) -> str:
        return EVENT_HANDLER_QUEUE.format(
            self.get_exchange_name(),
            self.get_routing_key(),
            self.dependant.name,
        )

    async def _handle_message(
            self,
            stack: AsyncExitStack,
            payload: Any,
            message: IncomingMessage,
    ) -> None:
        if isclass(self.payload_type) and issubclass(self.payload_type, BaseModel):
            try:
                payload = self.payload_type(**payload)
            except ValidationError as e:
                await message.ack()
                raise e

        params = await self.dependant.prepare_params(
            stack, {
                'payload': payload,
                'connection': self.connection,
                'context': self.context,
        })
        await self.dependant.call(**params)
        await message.ack()


class AmqpEventPublisher:
    def __init__(
            self,
            *,
            publish_model: Any = None,
    ):
        self.field = None
        if publish_model:
            self.field = create_publish_field(
                name='publish_model',
                type_=publish_model,
            )

    async def __call__(
            self,
            connection: Connection,
            context: ServiceContext,
    ) -> AsyncGenerator[Callable, None]:
        channel = await connection.channel()
        exchange = await channel.get_exchange(context.name)

        async def publish(routing_key: str, payload: Any):
            body = serialize_payload(
                field=self.field,
                payload_content=payload,
            )
            await exchange.publish(Message(body), routing_key)

        try:
            yield publish
        finally:
            await channel.close()



import logging

import click

from uservice.runner import ServiceRunner
from uservice.supervisors import ChangeReloader, Multiprocess


logger = logging.getLogger('uservice')


@click.command()
@click.argument("service")
@click.option("--reload", is_flag=True, default=False, help="Enable auto-reload.")
@click.option(
    "--workers",
    default=1,
    type=int,
    show_default=True,
    help="Number of worker processes. Not valid with --reload"
)
def run(
        service: str,
        *,
        reload: bool,
        workers: int,
):

    if workers > 1 and reload:
        logger.warning(
            "Can not have --reload and more than 1 workers at the same time."
        )

    runner = ServiceRunner(service_str=service)

    if reload:
        ChangeReloader(runner.run).run()

    else:
        Multiprocess(runner.run, workers).run()



from typing import Callable
from .dependencies import Dependant
from .contexts import ServiceContext


class Entrypoint:
    def __init__(
            self,
            *,
            context: ServiceContext,
            call: Callable,
    ):
        self.context = context
        self.dependant = Dependant(call=call)

    async def setup(self, *args, **kwargs) -> None:
        raise NotImplementedError

    async def start(self) -> None:
        raise NotImplementedError

    async def stop(self) -> None:
        raise NotImplementedError



import logging
import json

from contextlib import AsyncExitStack

from aio_pika import (
    Connection,
    Queue,
    Exchange,
    ExchangeType,
    Channel,
    IncomingMessage,
)

from uservice.entrypoints import Entrypoint


logger = logging.getLogger('uservice')

EVENT_HANDLER_QUEUE = 'uservice-{}-{}-{}'
RPC_QUEUE = 'rpc-{}'
RPC_REPLY_QUEUE = 'rpc-reply-{}-{}'
RPC_ROUTING_KEY = '{}.{}'


class AmqpConsumer(Entrypoint):
    async def setup(self, connection: Connection) -> None:
        self.connection = connection
        self.channel: Channel = await self.connection.channel()
        self.exchange: Exchange = await self.channel.declare_exchange(
            **self.get_exchange_settings()
        )
        self.queue: Queue = await self.channel.declare_queue(
            **self.get_queue_settings(),
        )
        await self.queue.bind(self.exchange, self.get_routing_key())

    async def start(self) -> None:
        self.consumer_tag = await self.queue.consume(self.handle_message)

    async def stop(self) -> None:
        await self.queue.cancel(self.consumer_tag)
        await self.channel.close()

    async def handle_message(self, message: IncomingMessage) -> None:
        async with AsyncExitStack() as stack:
            body = json.loads(message.body)
            await self._handle_message(stack, body, message)

    async def _handle_message(self, stack, body, message) -> None:
        raise NotImplementedError

    def get_exchange_name(self):
        raise NotImplementedError

    def get_routing_key(self):
        raise NotImplementedError

    def get_queue_name(self):
        raise NotImplementedError

    def get_queue_settings(self):
        return {
            'name': self.get_queue_name(),
            'durable': True,
        }

    def get_exchange_settings(self):
        return {
            'name': self.get_exchange_name(),
            'type': ExchangeType.TOPIC,
            'durable': True,
            'auto_delete': False,
        }



import logging
import os
import threading
import signal
import multiprocessing

import click

from pathlib import Path
from typing import Optional, List
from watchfiles import watch, PythonFilter
from .subprocess import get_subprocess, Supervisor


logger = logging.getLogger('uservice')


class ChangeReloader(Supervisor):
    def __init__(self, target, reload_dirs = None):
        super().__init__(target)
        if reload_dirs is None:
            self.reload_dirs = [Path(os.getcwd()).cwd()]
        else:
            self.reload_dirs = reload_dirs
        self.watcher = watch(
            *self.reload_dirs,
            watch_filter=PythonFilter(),
            stop_event=self.should_exit,
            yield_on_timeout=True,
        )
        self.reloader_name = "ChangeReloader"

    def loop(self):
        for changes in self:
            if changes:
                logger.warning(
                    "%s detected changes in %s. Reloading...",
                    self.reloader_name,
                    ", ".join(map(_display_path, changes)),
                )
                self.restart()

    def __iter__(self):
        return self

    def __next__(self) -> Optional[List[Path]]:
        return self.should_restart()

    def startup(self) -> None:
        super().startup()
        message = f"Started reloader process [{self.pid}] using {self.reloader_name}"
        color_message = "Started reloader process [{}] using {}".format(
            click.style(str(self.pid), fg="cyan", bold=True),
            click.style(str(self.reloader_name), fg="cyan", bold=True),
        )
        logger.info(message, extra={"color_message": color_message})
        self.start_process()

    def should_restart(self) -> Optional[List[Path]]:
        self.pause()
        changes = next(self.watcher)
        if changes:
            return {Path(c[1]) for c in changes}

        return None

    def restart(self) -> None:
        self.process.terminate()
        self.process.join()
        self.start_process()

    def start_process(self):
        self.process = get_subprocess(self.target)
        self.process.start()

    def pause(self) -> None:
        if self.should_exit.wait(0.25):
            raise StopIteration()

    def shutdown(self) -> None:
        self.process.terminate()  # pragma: py-win32
        self.process.join()

        message = "Stopping reloader process [{}]".format(str(self.pid))
        color_message = "Stopping reloader process [{}]".format(
            click.style(str(self.pid), fg="cyan", bold=True)
        )
        logger.info(message, extra={"color_message": color_message})


def _display_path(path: Path) -> str:
    try:
        return f"'{path.relative_to(Path.cwd())}'"
    except ValueError:
        return f"'{path}'"



import asyncio
import functools
import inspect
from typing import Callable, Dict, List, Any, Annotated
from contextlib import asynccontextmanager, AsyncExitStack
from pydantic.typing import get_origin, get_args


class Depends:
    def __init__(self, dependency: Callable):
        self.dependency = dependency


class Dependant:
    def __init__(self, *, call: Callable):
        self.call = call
        self.params = inspect.signature(call).parameters
        self.required_params = {}
        self.dependencies = {}
        for param_name, param in self.params.items():
            if isinstance(param.default, Depends):
                call = param.default.dependency
                self.dependencies[param_name] = Dependant(call=call)

            elif (
                    param.annotation is not inspect.Signature.empty
                    and get_origin(param.annotation) is Annotated
            ):
                annotation = get_args(param.annotation)[1]
                if isinstance(annotation, Depends):
                    call = annotation.dependency
                    self.dependencies[param_name] = Dependant(call=call)
            elif param.default is inspect.Signature.empty:
                self.required_params[param_name] = param

    @property
    def name(self):
        return self.call.__name__

    async def prepare_params(self, stack: AsyncExitStack, params: Dict[str, Any]):
        _params = {}
        for param in self.required_params:
            _params[param] = params[param]

        _params.update(
            await solve_dependencies(
                stack=stack,
                params=params,
                dependencies=self.dependencies,
            )
        )
        return _params


async def solve_dependencies(
        *,
        stack: AsyncExitStack,
        params: Dict[str, Any],
        dependencies: Dict[str, Dependant],
):
    deps = {}
    for kw, dependency in dependencies.items():
        dependency_params = await dependency.prepare_params(stack, params)
        if is_async_gen_callable(dependency.call):
            cm = asynccontextmanager(dependency.call)(**dependency_params)
            deps[kw] = await stack.enter_async_context(cm)
        elif is_async_callable(dependency.call):
            deps[kw] = await dependency.call(**dependency_params)
        else:
            deps[kw] = dependency.call(**dependency_params)

    return deps


def is_async_gen_callable(obj: Any) -> bool:
    return inspect.isasyncgenfunction(obj) or inspect.isasyncgenfunction(obj.__call__)


def is_async_callable(obj: Any) -> bool:
    while isinstance(obj, functools.partial):
        obj = obj.func

    return asyncio.iscoroutinefunction(obj) or (
        callable(obj) and asyncio.iscoroutinefunction(obj.__call__)
    )



import logging
import asyncio
import signal
import threading
from typing import Optional, NoReturn, Callable, Any, List
from aio_pika import Connection, connect
from .amqp.events import AmqpEventHandler
from .amqp.rpc import AmqpRpc
from .settings import get_settings, Settings
from .contexts import ServiceContext
from .entrypoints import Entrypoint


HANDLED_SIGNALS = (
    signal.SIGINT,  # Unix signal 2. Sent by Ctrl+C.
    signal.SIGTERM,  # Unix signal 15. Sent by `kill <pid>`.
)

logger = logging.getLogger('uservice')


class Service:
    def __init__(
            self,
            *,
            name: str,
            settings: Optional[Settings] = None,
    ):
        if settings is None:
            _settings = get_settings()

        self.entrypoints: List[Entrypoint] = []
        self.init_context(name, settings or _settings)

    def init_context(self, name: str, settings: Settings) -> None:
        self.context = ServiceContext(
            name=name,
            settings=settings,
        )

    def handle_exit(self, sig, frame) -> None:
        self.is_running = False

    async def run(self) -> None:
        connection = await connect(self.context.settings.amqp.get_url())
        async with connection:
            await self.setup(connection)
            await self.start()
            await self.loop()
            await self.stop()

    async def loop(self) -> None:
        while self.is_running:
            await asyncio.sleep(0.1)

    async def setup(self, connection: Connection) -> None:
        self.install_signal_handlers()
        for entrypoint in self.entrypoints:
            await entrypoint.setup(connection)

    async def start(self) -> None:
        for entrypoint in self.entrypoints:
            await entrypoint.start()

        self.is_running = True

    async def stop(self) -> None:
        for entrypoint in self.entrypoints:
            await entrypoint.stop()

    def event_handler(
            self,
            exchange: str,
            queue: str,
    ) -> Callable:
        def decorator(func: Callable) -> None:
            self.entrypoints.append(
                AmqpEventHandler(
                    context=self.context,
                    call=func,
                    exchange_name=exchange,
                    routing_key=queue,
                )
            )
            return func

        return decorator

    def rpc(
            self,
            *,
            response_model: Optional[Any] = None,
    ) -> Callable:
        def decorator(func: Callable, *args, **kwargs) -> None:
            self.entrypoints.append(
                AmqpRpc(
                    context=self.context,
                    call=func,
                    response_model=response_model,
                )
            )

        return decorator

    def install_signal_handlers(self) -> None:
        if threading.current_thread() is not threading.main_thread():
            # Signals can only be listened to from the main thread.
            return

        loop = asyncio.get_event_loop()

        try:
            for sig in HANDLED_SIGNALS:
                loop.add_signal_handler(sig, self.handle_exit, sig, None)
        except NotImplementedError:  # pragma: no cover
            # Windows
            for sig in HANDLED_SIGNALS:
                signal.signal(sig, self.handle_exit)



import click
import yaml

from uservice.importer import import_from_string


@click.command()
@click.argument("service")
def schema(
        service: str,
):
    loaded_service = import_from_string(service)
    schema = loaded_service.schema()
    with open('asyncapi.yaml', 'w') as f:
        yaml.dump(schema, f)



from functools import lru_cache
from typing import Optional
from pydantic import BaseSettings, AmqpDsn, Field, BaseModel


class AmqpSettings(BaseModel):
    url: AmqpDsn = 'amqp://guest:guest@localhost:5672/'
    host: Optional[str]
    port: Optional[int]
    user: Optional[str]
    password: Optional[str]
    tls: bool = False
    rpc_exchange: str = 'uservice-rpc'

    def get_url(self):
        if (
                self.host or
                self.port or
                self.user or
                self.password
        ):
            protocol = 'amqps' if self.tls else 'amqp'
            return f'{protocol}://{self.user}:{self.password}@{self.host}:{self.port}'

        return self.url


class AsyncAPISettings(BaseModel):
    version: str = '2.6.0'
    title: Optional[str]
    description: str = 'My Service'


class Settings(BaseSettings):
    communication_backend: str = 'amqp'
    amqp: AmqpSettings = AmqpSettings()
    asyncapi: AsyncAPISettings = AsyncAPISettings()

    class Config:
        env_nested_delimiter = '__'


@lru_cache()
def get_settings() -> Settings:
    return Settings()



import logging
import click

from .subprocess import Supervisor, get_subprocess

logger = logging.getLogger('uservice')


class Multiprocess(Supervisor):
    def __init__(self, target, workers):
        super().__init__(target)
        self.workers = workers
        self.processes = []

    def startup(self) -> None:
        super().startup()
        message = "Started parent process [{}]".format(str(self.pid))
        color_message = "Started parent process [{}]".format(
            click.style(str(self.pid), fg="cyan", bold=True)
        )
        logger.info(message, extra={"color_message": color_message})

        for _idx in range(self.workers):
            process = get_subprocess(
                target=self.target,
            )
            process.start()
            logger.info(f"Started worker process [{process.pid}]")
            self.processes.append(process)

    def loop(self):
        self.should_exit.wait()

    def shutdown(self) -> None:
        for process in self.processes:
            logger.info(f"Stopping worker process [{process.pid}]")
            process.terminate()
            process.join()

        message = "Stopping parent process [{}]".format(str(self.pid))
        color_message = "Stopping parent process [{}]".format(
            click.style(str(self.pid), fg="cyan", bold=True)
        )
        logger.info(message, extra={"color_message": color_message})



import json
import dataclasses
from typing import Type, Any, Optional
from pydantic import BaseConfig, BaseModel, ValidationError
from pydantic.fields import ModelField, FieldInfo


def create_publish_field(
    name: str,
    type_: Type[Any],
) -> ModelField:
    """
    Create a new publish field. Raises if type_ is invalid.
    """

    try:
        return ModelField(
            name=name,
            type_=type_,
            class_validators={},
            model_config=BaseConfig,
            field_info=FieldInfo()
        )
    except RuntimeError:
        raise Exception(
            "Invalid args for publish field! Hint: "
            f"check that {type_} is a valid Pydantic field type. "
            "If you are using a return type annotation that is not a valid Pydantic "
            "field (e.g. Union[Response, dict, None]) you can disable generating the "
            "publish model from the type annotation with not setting publish_model "
            "in EventPublisher()"
        ) from None


def serialize_payload(
        *,
        field: Optional[ModelField] = None,
        payload_content: Any,
):
    payload_content = _prepare_payload_content(payload_content)
    if not field:
        return _encode(payload_content)

    body, errors = field.validate(payload_content, {}, loc='payload')
    if errors:
        raise ValidationError(errors, field)

    return _encode(body.dict())


def _prepare_payload_content(
    payload: Any,
) -> Any:
    if isinstance(payload, BaseModel):
        read_with_orm_mode = getattr(payload.__config__, "read_with_orm_mode", None)
        if read_with_orm_mode:
            return payload

        return payload.dict(by_alias=True)
    elif isinstance(payload, list):
        return [
            _prepare_payload_content(item)
            for item in payload
        ]

    elif isinstance(payload, dict):
        return {
            k: _prepare_payload_content(v)
            for k, v in payload.items()
        }

    elif dataclasses.is_dataclass(payload):
        return dataclasses.asdict(payload)

    return payload


def _encode(data: Any) -> bytes:
    return json.dumps(data).encode()



import sys
import os
import multiprocessing
import threading
import signal
from typing import Optional

spawn = multiprocessing.get_context("spawn")

HANDLED_SIGNALS = (
    signal.SIGINT,  # Unix signal 2. Sent by Ctrl+C.
    signal.SIGTERM,  # Unix signal 15. Sent by `kill <pid>`.
)


class Supervisor:
    def __init__(self, target):
        self.target = target
        self.pid = os.getpid()
        self.should_exit = threading.Event()

    def signal_handler(self, sig, frame) -> None:
        self.should_exit.set()

    def run(self):
        self.startup()
        self.loop()
        self.shutdown()

    def startup(self):
        for sig in HANDLED_SIGNALS:
            signal.signal(sig, self.signal_handler)

    def shutdown(self):
        raise NotImplementedError


def get_subprocess(target):
    stdin_fileno: Optional[int]
    try:
        stdin_fileno = sys.stdin.fileno()
    except OSError:
        stdin_fileno = None
    kwargs = {
        "target": target,
        "stdin_fileno": stdin_fileno,
    }
    return spawn.Process(target=subprocess_started, kwargs=kwargs)


def subprocess_started(
        target,
        stdin_fileno: Optional[int],
) -> None:
    if stdin_fileno is not None:
        sys.stdin = os.fdopen(stdin_fileno)

    target()



from typing import Callable

from uservice.entrypoints import Entrypoint
from uservice.contexts import ServiceContext


class RedisEventHandler(Entrypoint):
    def __init__(
            self,
            *,
            context: ServiceContext,
            call: Callable,
            source: str,
            event: str,
    ):
        super().__init__(context=context, call=call)
        self.source = source
        self.event = event



from __future__ import annotations

import asyncio
import uuid
import json

from typing import Callable, Any, Optional, AsyncGenerator
from contextlib import AsyncExitStack

from aio_pika import IncomingMessage, Connection, Message, Exchange, Channel

from uservice.contexts import ServiceContext
from uservice.utils import serialize_payload
from .consumer import AmqpConsumer


RPC_QUEUE = 'rpc-{}'
RPC_REPLY_QUEUE = 'rpc-reply-{}-{}'
RPC_ROUTING_KEY = '{}.{}'


class AmqpRpc(AmqpConsumer):
    def __init__(
            self,
            *,
            context: ServiceContext,
            call: Callable,
            response_model: Optional[Any],
    ):
        super().__init__(context=context, call=call)
        self.response_model = response_model

    def get_exchange_name(self) -> str:
        return self.context.settings.amqp.rpc_exchange

    def get_routing_key(self) -> str:
        return RPC_ROUTING_KEY.format(self.context.name, self.dependant.name)

    def get_queue_name(self) -> str:
        return RPC_QUEUE.format(self.context.name)

    async def _handle_message(
            self,
            stack: AsyncExitStack,
            body: Any,
            message: IncomingMessage,
    ) -> None:
        reply_to = message.reply_to
        correlation_id = message.correlation_id
        params = await self.dependant.prepare_params(
            stack,
            {
                'connection': self.connection,
                'context': self.context,
                **body['kwargs'],
            })
        result = await self.dependant.call(**params)
        body = serialize_payload(
            field=None,
            payload_content=result,
        )
        response = Message(
            body=body,
            correlation_id=correlation_id,
        )
        await self.exchange.publish(response, reply_to)
        await message.ack()


class AmqpRpcProxy:
    def __init__(
            self,
            *,
            target_service: str,
    ):
        self.target_service = target_service

    async def __call__(
            self,
            connection: Connection,
            context: ServiceContext,
    ) -> AsyncGenerator[ServiceProxy, None]:
        proxy: ServiceProxy
        try:
            channel = await connection.channel()
            exchange = await channel.get_exchange(context.settings.amqp.rpc_exchange)
            proxy = ServiceProxy(
                channel=channel,
                exchange=exchange,
                service_name=context.name,
                target_service=self.target_service,
            )
            yield proxy
        finally:
            await proxy.stop()
            await channel.close()


class ServiceProxy:
    def __init__(
            self,
            *,
            channel: Channel,
            exchange: Exchange,
            service_name: str,
            target_service: str,
    ):

        self.channel = channel
        self.exchange = exchange
        self.service_name = service_name
        self.target_service = target_service

    def __getattr__(self, name) -> MethodProxy:
        self.proxy = MethodProxy(
            channel=self.channel,
            exchange=self.exchange,
            service_name=self.service_name,
            target_service=self.target_service,
            method_name=name,
        )
        return self.proxy

    async def stop(self) -> None:
        await self.proxy.stop()


class MethodProxy:
    def __init__(
            self,
            *,
            channel: Channel,
            exchange: Exchange,
            service_name: str,
            target_service: str,
            method_name: str,
    ):
        self.channel = channel
        self.exchange = exchange
        self.service_name = service_name
        self.target_service = target_service
        self.method_name = method_name

    async def __call__(self, **kwargs) -> Any:
        reply_routing_key = await self.setup_reply_queue()
        body = serialize_payload(
            field=None,
            payload_content={'kwargs': kwargs},
        )
        self.correlation_id = str(uuid.uuid4())
        message = Message(
            body=body,
            reply_to=reply_routing_key,
            correlation_id=self.correlation_id,
        )
        self.response: asyncio.Future = asyncio.Future()
        await self.exchange.publish(message, RPC_ROUTING_KEY.format(self.target_service, self.method_name))

        await self.response
        return self.response.result()

    async def setup_reply_queue(self) -> str:
        reply_routing_key = str(uuid.uuid4())
        reply_queue_name = RPC_REPLY_QUEUE.format(self.service_name, reply_routing_key)
        self.reply_queue = await self.channel.declare_queue(
            reply_queue_name,
            auto_delete=True,
        )
        await self.reply_queue.bind(self.exchange, reply_routing_key)
        self.consumer_tag = await self.reply_queue.consume(self.handle_message)
        return reply_routing_key

    async def handle_message(self, message: IncomingMessage) -> None:
        body = json.loads(message.body)
        if message.correlation_id == self.correlation_id:
            self.response.set_result(body)

        await message.ack()

    async def stop(self) -> None:
        await self.reply_queue.cancel(self.consumer_tag)
        await self.reply_queue.delete()
